/*
 * deviced-shutdown: deviced's replacement for systemd-shutdown binary
 *
 * This binary is supposed to be called on Tizen system by systemd
 * manager during shutdown, as result of eg. dbus call to Reboot
 * method, test via:
 *
 *  # dbus-send --print-reply --system --dest=org.freedesktop.systemd1 --type=method_call /org/freedesktop/systemd1 org.freedesktop.systemd1.Manager.(Reboot|PowerOff)
 *
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <linux/reboot.h>
#include <signal.h>
#include <sys/reboot.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "dd-deviced.h"
#include "core/log.h"
#include "core/common.h"

enum {
	CMD_REBOOT,
	CMD_POWEROFF,
	CMD_HALT = CMD_POWEROFF,
	CMD_EXIT,
};

#define UMOUNT_RW_PATH_SDCARD		"/opt/media/SDCardA1"
#define UMOUNT_RW_PATH_USER		"/opt/usr"
#define UMOUNT_RW_PATH_SYSTEM		"/opt"
#define MAX_UMOUNT_KILL_RETRY		4

char *progname;

static int usage(void)
{
	printf("%s: Invalid command specified (supported are: reboot, halt|poweroff, exit)\n", progname);
	return 0;
}

static void umount_partitions(void)
{
	umount_partition_by_kill(UMOUNT_RW_PATH_SDCARD, MAX_UMOUNT_KILL_RETRY);
	umount_partition_by_kill(UMOUNT_RW_PATH_USER, MAX_UMOUNT_KILL_RETRY);
	umount_partition_by_kill(UMOUNT_RW_PATH_SYSTEM, MAX_UMOUNT_KILL_RETRY);
}

static void sigchld(int signo)
{
	int wstatus;
	int r;

	do
		r = waitpid(-1, &wstatus, WNOHANG);
	while (r > 0);
}

static void autoboot(void)
{
	static char buf[4096] = {0,};
	int fd = open("/run/reboot", O_RDONLY); // XXX POWER_FLAG_REBOOT
	int ret;

	if (fd >= 0) {
		ret = read(fd, buf, sizeof(buf) - 1);
		if (ret < 0) {
			ret = -errno;
			printf("%s: fail to read reboot data(%d)\n", progname, ret);
		}
		close(fd);
	}

	printf("%s: reboot param is <%s>\n", progname, buf);
	syscall(SYS_reboot, LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2,
		buf[0] != 0 ? LINUX_REBOOT_CMD_RESTART2 : LINUX_REBOOT_CMD_RESTART,
		&buf);
}

static void handle_exit(void)
{
	/* systemd default behviour for handling exit in system manager
	 * is to treat is as poweroff.  Exiting from PID1 would cause
	 * system to panic, so it's not really "supported.  Poweroff is
	 * sane default, but it's also implementation detail and something
	 * user should depend on.
	 *
	 * In Tizen we put shutdown binary to sleep, rather than powering
	 * it off.  This functionaliy has been requested by test team.
	 */
	for (;;)
		pause();
}

int main(int ac, char *av[])
{
	enum {
		ARG_LOGLEVEL,
		ARG_LOGTARGET,
		ARG_LOGCOLOR,
		ARG_LOGLOCATION,
		ARG_EXITCODE
	};
	static struct option opts[] = {
		{"log-level", required_argument, NULL, ARG_LOGLEVEL},
		{"log-target", required_argument, NULL, ARG_LOGTARGET},
		{"log-color", optional_argument, NULL, ARG_LOGCOLOR},
		{"log-location", optional_argument, NULL, ARG_LOGLOCATION},
		{"exit-code", required_argument, NULL, ARG_EXITCODE},
		{},
	};

	progname = av[0];

	// Ensure messages are printed to console as fast as possible.
	// Otherwise we risk that in case of error nothing will be printed
	// as last step is reboot.
	setvbuf(stdout, NULL, _IONBF, 0);

	int c;
	while ((c = getopt_long(ac, av, "", opts, NULL)) >= 0)
		; // not supported for now

	int cmd;
	char *cmd_name = av[optind];
	if (optind >= ac || !cmd_name)
		return usage();

	if (strcmp(cmd_name, "reboot") == 0)
		cmd = CMD_REBOOT;
	else if (strcmp(cmd_name, "halt") == 0 || strcmp(cmd_name, "poweroff") == 0)
		cmd = CMD_POWEROFF;
	else if (strcmp(cmd_name, "exit") == 0)
		cmd = CMD_EXIT;
	else
		// XXX this is critical failure - we are pid1 already
		// and can't perform requested action - exit below
		// will cause system to panic() and, hopefully, reboot
		// (depends on sysctls)
		return usage();

	printf("%s: executing %s action\n", progname, cmd_name);

	(void)signal(SIGCHLD, sigchld);
	(void)signal(SIGHUP, SIG_IGN);
	(void)signal(SIGTERM, SIG_IGN);

	printf("%s: unmounting rw partitions\n", progname);

	umount_partitions();

	printf("%s: rw partitions have been unmounted\n", progname);

	sync();

	printf("%s: performing %s\n", progname, cmd_name);

	if (cmd == CMD_REBOOT)
		autoboot();
	else if (cmd == CMD_EXIT)
		handle_exit();
	else
		reboot(RB_POWER_OFF);

	return 1;
}
